#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains methods used by the paasta client to wait for deployment
of a docker image to a cluster.instance.
"""
import asyncio
import logging
from typing import Optional

from paasta_tools.cli.cmds.mark_for_deployment import NoSuchCluster
from paasta_tools.cli.cmds.mark_for_deployment import report_waiting_aborted
from paasta_tools.cli.cmds.mark_for_deployment import wait_for_deployment
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import NoSuchService
from paasta_tools.cli.utils import validate_git_sha
from paasta_tools.cli.utils import validate_given_deploy_groups
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.cli.utils import validate_short_git_sha
from paasta_tools.remote_git import list_remote_refs
from paasta_tools.remote_git import LSRemoteException
from paasta_tools.utils import _log
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeploymentVersion
from paasta_tools.utils import get_git_url
from paasta_tools.utils import get_latest_deployment_tag
from paasta_tools.utils import list_services
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import TimeoutError

DEFAULT_DEPLOYMENT_TIMEOUT = 3600  # seconds


log = logging.getLogger(__name__)


class VersionError(Exception):
    pass


class DeployGroupError(Exception):
    pass


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        "wait-for-deployment",
        help="Wait a service to be deployed to deploy_group",
        description=(
            "'paasta wait-for-deployment' waits for a previously marked for "
            "deployment service to be deployed to deploy_group."
        ),
        epilog=(
            "Note: Access and credentials to the Git repo of a service "
            "are required for this command to work."
        ),
    )
    list_parser.add_argument(
        "-u",
        "--git-url",
        help=(
            "Git url for service. Defaults to the normal git URL for " "the service."
        ),
        default=None,
    )
    list_parser.add_argument(
        "-c",
        "-k",
        "--commit",
        help="Git sha to wait for deployment",
        required=True,
        type=validate_short_git_sha,
    )
    list_parser.add_argument(
        "-i",
        "--image-version",
        help="Extra version metadata to mark for deployment",
        required=False,
        default=None,
    )
    list_parser.add_argument(
        "-l",
        "--deploy-group",
        help="deploy group (e.g. cluster1.canary, cluster2.main).",
        required=True,
    ).completer = lazy_choices_completer(list_deploy_groups)
    list_parser.add_argument(
        "-s",
        "--service",
        help="Name of the service which you wish to wait for deployment. "
        'Leading "services-" will be stripped.',
        required=True,
    ).completer = lazy_choices_completer(list_services)
    list_parser.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        type=int,
        default=DEFAULT_DEPLOYMENT_TIMEOUT,
        help=(
            "Time in seconds to wait for paasta to deploy the service. "
            "If the timeout is exceeded we return 1. "
            "Default is %(default)s seconds."
        ),
    )
    list_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    list_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help="Print out more output.",
    )
    list_parser.add_argument(
        "--polling-interval",
        dest="polling_interval",
        type=float,
        default=None,
        help="How long to wait between each time we check to see if an instance is done deploying.",
    )
    list_parser.add_argument(
        "--diagnosis-interval",
        dest="diagnosis_interval",
        type=float,
        default=None,
        help="How long to wait between diagnoses of why the bounce isn't done.",
    )
    list_parser.add_argument(
        "--time-before-first-diagnosis",
        dest="time_before_first_diagnosis",
        type=float,
        default=None,
        help="Wait this long before trying to diagnose why the bounce isn't done.",
    )

    list_parser.set_defaults(command=paasta_wait_for_deployment)


def get_latest_marked_version(
    git_url: str, deploy_group: str
) -> Optional[DeploymentVersion]:
    """Return the latest marked for deployment version or None"""
    # TODO: correct this function for new tag format
    refs = list_remote_refs(git_url)
    _, sha, image_version = get_latest_deployment_tag(refs, deploy_group)
    if sha:
        return DeploymentVersion(sha=sha, image_version=image_version)
    # We did not find a ref for this deploy group
    return None


def validate_version_is_latest(
    version: DeploymentVersion, git_url: str, deploy_group: str, service: str
):
    """Verify if the requested version  is the latest marked for deployment.

    Raise exception when the provided version is not the latest
    marked for deployment in 'deploy_group' for 'service'.
    """
    try:
        marked_version = get_latest_marked_version(git_url, deploy_group)
    except LSRemoteException as e:
        print(
            "Error talking to the git server: {}\n"
            "It is not possible to verify that {} is marked for deployment in {}, "
            "but I assume that it is marked and will continue waiting..".format(
                e, version, deploy_group
            )
        )
        return
    if marked_version is None:
        raise VersionError(
            "ERROR: Nothing is marked for deployment "
            "in {} for {}".format(deploy_group, service)
        )
    if version != marked_version:
        raise VersionError(
            "ERROR: The latest version marked for "
            "deployment in {} is {}".format(deploy_group, marked_version)
        )


def validate_deploy_group(deploy_group: str, service: str, soa_dir: str):
    """Validate deploy_group.

    Raise exception if the specified deploy group is not used anywhere.
    """
    in_use_deploy_groups = list_deploy_groups(service=service, soa_dir=soa_dir)
    _, invalid_deploy_groups = validate_given_deploy_groups(
        in_use_deploy_groups, [deploy_group]
    )

    if len(invalid_deploy_groups) == 1:
        raise DeployGroupError(
            "ERROR: These deploy groups are not currently "
            "used anywhere: {}.\n"
            "You probably need one of these in-use deploy "
            "groups?:\n   {}".format(
                ",".join(invalid_deploy_groups), ",".join(in_use_deploy_groups)
            )
        )


def paasta_wait_for_deployment(args):
    """Wrapping wait_for_deployment"""
    if args.verbose:
        log.setLevel(level=logging.DEBUG)
    else:
        log.setLevel(level=logging.INFO)

    service = args.service
    if service and service.startswith("services-"):
        service = service.split("services-", 1)[1]

    if args.git_url is None:
        args.git_url = get_git_url(service=service, soa_dir=args.soa_dir)

    args.commit = validate_git_sha(sha=args.commit, git_url=args.git_url)

    version = DeploymentVersion(sha=args.commit, image_version=args.image_version)

    try:
        validate_service_name(service, soa_dir=args.soa_dir)
        validate_deploy_group(args.deploy_group, service, args.soa_dir)
        validate_version_is_latest(version, args.git_url, args.deploy_group, service)
    except (VersionError, DeployGroupError, NoSuchService) as e:
        print(PaastaColors.red(f"{e}"))
        return 1

    try:
        asyncio.run(
            wait_for_deployment(
                service=service,
                deploy_group=args.deploy_group,
                git_sha=args.commit,
                image_version=args.image_version,
                soa_dir=args.soa_dir,
                timeout=args.timeout,
                polling_interval=args.polling_interval,
                diagnosis_interval=args.diagnosis_interval,
                time_before_first_diagnosis=args.time_before_first_diagnosis,
            )
        )
        _log(
            service=service,
            component="deploy",
            line=(f"Deployment of {version} for {args.deploy_group} complete"),
            level="event",
        )

    except (KeyboardInterrupt, TimeoutError, NoSuchCluster):
        report_waiting_aborted(service, args.deploy_group)
        return 1

    return 0



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Sequence
from typing import Tuple

from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet

from paasta_tools.kubernetes.application.controller_wrappers import Application
from paasta_tools.kubernetes.application.controller_wrappers import DeploymentWrapper
from paasta_tools.kubernetes.application.controller_wrappers import StatefulSetWrapper
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import paasta_prefixed

log = logging.getLogger(__name__)


def is_valid_application(deployment: V1Deployment):
    is_valid = True
    missing = []
    for attr in ["service", "instance", "git_sha", "config_sha"]:
        prefixed_attr = paasta_prefixed(attr)
        if prefixed_attr not in deployment.metadata.labels:
            is_valid = False
            missing.append(prefixed_attr)
    if missing:
        log.warning(
            f"deployment/{deployment.metadata.name} in "
            f"namespace/{deployment.metadata.namespace} "
            f"is missing following labels: {missing}"
        )
    return is_valid


def list_paasta_managed_deployments(
    kube_client: KubeClient, **kwargs
) -> Dict[Tuple[str, str], List[Application]]:
    deployments: Dict[Tuple[str, str], List[Application]] = {}
    for deployment in kube_client.deployments.list_deployment_for_all_namespaces(
        label_selector=paasta_prefixed("managed"), **kwargs
    ).items:
        if is_valid_application(deployment):
            application = DeploymentWrapper(deployment)
            service = application.kube_deployment.service
            instance = application.kube_deployment.instance
            if deployments.get((service, instance), None):
                deployments[(service, instance)].append(application)
            else:
                deployments[(service, instance)] = [application]
    return deployments


def list_paasta_managed_stateful_sets(
    kube_client: KubeClient, **kwargs
) -> Dict[Tuple[str, str], List[Application]]:
    deployments: Dict[Tuple[str, str], List[Application]] = {}
    for deployment in kube_client.deployments.list_stateful_set_for_all_namespaces(
        label_selector=paasta_prefixed("managed"), **kwargs
    ).items:
        if is_valid_application(deployment):
            application = StatefulSetWrapper(deployment)
            service = application.kube_deployment.service
            instance = application.kube_deployment.instance
            if deployments.get((service, instance), None):
                deployments[(service, instance)].append(application)
            else:
                deployments[(service, instance)] = [application]
    return deployments


def list_all_applications(
    kube_client: KubeClient, application_types: Sequence[Any]
) -> Dict[Tuple[str, str], List[Application]]:
    """
    List all applications in the cluster of the types from application_types.
    Only applications with complete set of labels are included (See is_valid_application()).
    :param kube_client:

    :param application_types:  types of applications
    :return: A mapping from (service, instance) to application
    """
    apps: Dict[Tuple[str, str], List[Application]] = {}
    for application_type in application_types:
        if application_type == V1Deployment:
            apps = {**apps, **list_paasta_managed_deployments(kube_client)}
        elif application_type == V1StatefulSet:
            apps.update(list_paasta_managed_stateful_sets(kube_client))
    return apps



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import math
from typing import Callable
from typing import Collection
from typing import Dict
from typing import Sequence
from typing import Set

from mypy_extensions import Arg
from mypy_extensions import DefaultArg
from mypy_extensions import TypedDict


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())
logging.getLogger("requests").setLevel(logging.WARNING)

ZK_LOCK_CONNECT_TIMEOUT_S = 10.0  # seconds to wait to connect to zookeeper
ZK_LOCK_PATH = "/bounce"
WAIT_CREATE_S = 3
WAIT_DELETE_S = 5

BounceMethodConfigDict = TypedDict("BounceMethodConfigDict", {"instances": int})

BounceMethodResult = TypedDict(
    "BounceMethodResult", {"create_app": bool, "tasks_to_drain": Set}
)

BounceMethod = Callable[
    [
        Arg(BounceMethodConfigDict, "new_config"),
        Arg(bool, "new_app_running"),
        Arg(Collection, "happy_new_tasks"),
        Arg(Sequence, "old_non_draining_tasks"),
        DefaultArg(float, "margin_factor"),
    ],
    BounceMethodResult,
]


_bounce_method_funcs: Dict[str, BounceMethod] = {}


def register_bounce_method(name: str) -> Callable[[BounceMethod], BounceMethod]:
    """Returns a decorator that registers that bounce function at a given name
    so get_bounce_method_func can find it."""

    def outer(bounce_func: BounceMethod):
        _bounce_method_funcs[name] = bounce_func
        return bounce_func

    return outer


def get_bounce_method_func(name) -> BounceMethod:
    return _bounce_method_funcs[name]


def list_bounce_methods() -> Collection[str]:
    return _bounce_method_funcs.keys()


@register_bounce_method("brutal")
def brutal_bounce(
    new_config: BounceMethodConfigDict,
    new_app_running: bool,
    happy_new_tasks: Collection,
    old_non_draining_tasks: Sequence,
    margin_factor=1.0,
) -> BounceMethodResult:
    """Pays no regard to safety. Starts the new app if necessary, and kills any
    old ones. Mostly meant as an example of the simplest working bounce method,
    but might be tolerable for some services.

    :param new_config: The configuration dictionary representing the desired new app.
    :param new_app_running: Whether there is an app in Marathon with the same ID as the new config.
    :param happy_new_tasks: Set of MarathonTasks belonging to the new application that are considered healthy and up.
    :param old_non_draining_tasks: A sequence of tasks not belonging to the new version. Tasks should be ordered from
                                   most desirable to least desirable.
    :param margin_factor: the multiplication factor used to calculate the number of instances to be drained
                          when the crossover method is used.
    :return: A dictionary representing the desired bounce actions and containing the following keys:
              - create_app: True if we should start the new Marathon app, False otherwise.
              - tasks_to_drain: a set of task objects which should be drained and killed. May be empty.
    """
    return {
        "create_app": not new_app_running,
        "tasks_to_drain": set(old_non_draining_tasks),
    }


@register_bounce_method("upthendown")
def upthendown_bounce(
    new_config: BounceMethodConfigDict,
    new_app_running: bool,
    happy_new_tasks: Collection,
    old_non_draining_tasks: Sequence,
    margin_factor=1.0,
) -> BounceMethodResult:
    """Starts a new app if necessary; only kills old apps once all the requested tasks for the new version are running.

    See the docstring for brutal_bounce() for parameters and return value.
    """
    if new_app_running and len(happy_new_tasks) == new_config["instances"]:
        return {"create_app": False, "tasks_to_drain": set(old_non_draining_tasks)}
    else:
        return {"create_app": not new_app_running, "tasks_to_drain": set()}


@register_bounce_method("crossover")
def crossover_bounce(
    new_config: BounceMethodConfigDict,
    new_app_running: bool,
    happy_new_tasks: Collection,
    old_non_draining_tasks: Sequence,
    margin_factor=1.0,
) -> BounceMethodResult:
    """Starts a new app if necessary; slowly kills old apps as instances of the new app become happy.

    See the docstring for brutal_bounce() for parameters and return value.
    """

    assert margin_factor > 0
    assert margin_factor <= 1

    needed_count = max(
        int(math.ceil(new_config["instances"] * margin_factor)) - len(happy_new_tasks),
        0,
    )

    return {
        "create_app": not new_app_running,
        "tasks_to_drain": set(old_non_draining_tasks[needed_count:]),
    }


@register_bounce_method("downthenup")
def downthenup_bounce(
    new_config: BounceMethodConfigDict,
    new_app_running: bool,
    happy_new_tasks: Collection,
    old_non_draining_tasks: Sequence,
    margin_factor=1.0,
) -> BounceMethodResult:
    """Stops any old apps and waits for them to die before starting a new one.

    See the docstring for brutal_bounce() for parameters and return value.
    """
    return {
        "create_app": not old_non_draining_tasks and not new_app_running,
        "tasks_to_drain": set(old_non_draining_tasks),
    }


@register_bounce_method("down")
def down_bounce(
    new_config: BounceMethodConfigDict,
    new_app_running: bool,
    happy_new_tasks: Collection,
    old_non_draining_tasks: Sequence,
    margin_factor=1.0,
) -> BounceMethodResult:
    """
    Stops old apps, doesn't start any new apps.
    Used for the graceful_app_drain script.
    """
    return {"create_app": False, "tasks_to_drain": set(old_non_draining_tasks)}


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import TypedDict
from typing import Union

import service_configuration_lib

from paasta_tools.kubernetes_tools import KubernetesDeploymentConfigDict
from paasta_tools.kubernetes_tools import sanitised_cr_name
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_service_instance_config
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.vitesscluster_tools import get_formatted_environment_variables
from paasta_tools.vitesscluster_tools import KVEnvVar
from paasta_tools.vitesscluster_tools import KVEnvVarValueFrom
from paasta_tools.vitesscluster_tools import RequestsDict
from paasta_tools.vitesscluster_tools import ResourceConfigDict
from paasta_tools.vitesscluster_tools import VitessDeploymentConfig
from paasta_tools.vitesscluster_tools import VitessDeploymentConfigDict


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


KUBERNETES_NAMESPACE = "paasta-vitessclusters"


# Global variables
TOPO_IMPLEMENTATION = "zk2"
TOPO_GLOBAL_ROOT = "/vitess-paasta/global"

VTGATE_EXTRA_ENV = {
    "VAULT_ROLEID": {
        "secretKeyRef": {
            "name": "paasta-vitessclusters-secret-vitess-k8s-vault-vtgate-approle-roleid",
            "key": "vault-vtgate-approle-roleid",
        }
    },
    "VAULT_SECRETID": {
        "secretKeyRef": {
            "name": "paasta-vitessclusters-secret-vitess-k8s-vault-vtgate-approle-secretid",
            "key": "vault-vtgate-approle-secretid",
        }
    },
}


class GatewayConfigDict(TypedDict, total=False):
    affinity: Dict[str, Any]
    extraEnv: List[Union[KVEnvVar, KVEnvVarValueFrom]]
    extraFlags: Dict[str, str]
    extraLabels: Dict[str, str]
    replicas: int
    resources: Dict[str, Any]
    annotations: Mapping[str, Any]


class VitessCellConfigDict(VitessDeploymentConfigDict, total=False):
    name: str
    allCells: List[str]
    globalLockserver: Dict[str, str]
    gateway: GatewayConfigDict


def get_cell_config(
    cell: str,
    images: Dict[str, str],
    allCells: List[str],
    global_lock_server: Dict[str, str],
    region: str,
    vtgate_resources: ResourceConfigDict,
    env: List[Union[KVEnvVar, KVEnvVarValueFrom]],
    labels: Dict[str, str],
    node_affinity: dict,
    annotations: Mapping[str, Any],
) -> VitessCellConfigDict:
    """
    get vtgate config
    """
    replicas = vtgate_resources.get("replicas")
    requests = vtgate_resources.get(
        "requests", RequestsDict(cpu="100m", memory="256Mi")
    )
    environment_overrides: Dict[str, Any] = {
        "VAULT_ADDR": f"https://vault-dre.{region}.yelpcorp.com:8200",
        "VAULT_CACERT": f"/etc/vault/all_cas/acm-privateca-{region}.crt",
    }
    environment_overrides.update(VTGATE_EXTRA_ENV)
    updated_vtgate_extra_env = (
        get_formatted_environment_variables(environment_overrides) + env
    )

    config = VitessCellConfigDict(
        name=cell,
        images=images,
        allCells=allCells,
        globalLockserver=global_lock_server,
        gateway=GatewayConfigDict(
            affinity={"nodeAffinity": node_affinity},
            extraEnv=updated_vtgate_extra_env,
            extraFlags={
                "mysql_auth_server_impl": "vault",
                "mysql_auth_vault_addr": f"https://vault-dre.{region}.yelpcorp.com:8200",
                "mysql_auth_vault_path": "secrets/vitess/vt-gate/vttablet_credentials.json",
                "mysql_auth_vault_tls_ca": f"/etc/vault/all_cas/acm-privateca-{region}.crt",
                "mysql_auth_vault_ttl": "60s",
            },
            extraLabels=labels,
            replicas=replicas,
            resources={
                "requests": requests,
                "limits": requests,
            },
            annotations=annotations,
        ),
    )
    return config


class VitessCellInstanceConfigDict(KubernetesDeploymentConfigDict, total=False):
    cell: str
    cells: List[str]
    zk_address: str
    vtgate_resources: ResourceConfigDict
    images: Dict[str, str]


class VitessCellConfig(VitessDeploymentConfig):
    config_dict: VitessCellInstanceConfigDict

    config_filename_prefix = "vitesscell"

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: VitessCellConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,  # superregion
            instance=instance,  # host-1
            service=service,  # vitess
            soa_dir=soa_dir,
            config_dict=config_dict,
            branch_dict=branch_dict,
        )

    def get_images(self) -> Dict[str, str]:
        vitess_images = self.config_dict.get(
            "images", load_system_paasta_config().get_vitess_images()
        )
        return {
            "vtctld": vitess_images["vtctld_image"],
            "vtadmin": vitess_images["vtadmin_image"],
            "vtgate": vitess_images["vtgate_image"],
            "vttablet": vitess_images["vttablet_image"],
        }

    def get_global_lock_server(self) -> Dict[str, str]:
        zk_address = self.config_dict.get("zk_address")
        return {
            "implementation": TOPO_IMPLEMENTATION,
            "address": zk_address,
            "rootPath": TOPO_GLOBAL_ROOT,
        }

    def get_vitess_cell_config(self) -> VitessCellConfigDict:
        cell = self.config_dict.get("cell")
        all_cells = self.config_dict.get("cells")
        images = self.get_images()
        global_lock_server = self.get_global_lock_server()
        region = self.get_region()
        vtgate_resources = self.config_dict.get("vtgate_resources")

        formatted_env = self.get_env_variables()
        labels = self.get_labels()
        node_affinity = self.get_vitess_node_affinity()
        annotations = self.get_annotations()

        return get_cell_config(
            cell,
            images,
            all_cells,
            global_lock_server,
            region,
            vtgate_resources,
            formatted_env,
            labels,
            node_affinity,
            annotations,
        )

    def validate(
        self,
        params: List[str] = [
            "cpus",
            "security",
            "dependencies_reference",
            "deploy_group",
        ],
    ) -> List[str]:
        # Use InstanceConfig to validate shared config keys like cpus and mem
        # TODO: add mem back to this list once we fix PAASTA-15582 and
        # move to using the same units as flink/marathon etc.
        error_msgs = super().validate(params=params)

        if error_msgs:
            name = self.get_instance()
            return [f"{name}: {msg}" for msg in error_msgs]
        else:
            return []


def load_vitess_cell_instance_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessCellConfig:
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    instance_config = load_service_instance_config(
        service, instance, "vitesscell", cluster, soa_dir=soa_dir
    )
    general_config = deep_merge_dictionaries(
        overrides=instance_config, defaults=general_config
    )

    branch_dict: Optional[BranchDictV2] = None
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = VitessCellConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    vitess_cell_config = VitessCellConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )

    return vitess_cell_config


def load_vitess_cell_instance_configs(
    service: str,
    instance: str,
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessCellConfigDict:
    vitess_cell_instance_configs = load_vitess_cell_instance_config(
        service, instance, cluster, soa_dir=soa_dir
    ).get_vitess_cell_config()
    return vitess_cell_instance_configs


# TODO: read this from CRD in service configs
def cr_id(service: str, instance: str) -> Mapping[str, str]:
    return dict(
        group="planetscale.com",
        version="v2",
        namespace=KUBERNETES_NAMESPACE,
        plural="vitesscells",
        name=sanitised_cr_name(service, instance),
    )



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeploymentVersion
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import NoDeploymentsAvailable


def get_currently_deployed_sha(service, deploy_group, soa_dir=DEFAULT_SOA_DIR):
    """Tries to determine the currently deployed sha for a service and deploy_group,
    returns None if there isn't one ready yet"""
    try:
        deployments = load_v2_deployments_json(service=service, soa_dir=soa_dir)
        return deployments.get_git_sha_for_deploy_group(deploy_group=deploy_group)
    except NoDeploymentsAvailable:
        return None


def get_currently_deployed_version(
    service, deploy_group, soa_dir=DEFAULT_SOA_DIR
) -> Optional[DeploymentVersion]:
    """Tries to determine the currently deployed version for a service and deploy_group,
    returns None if there isn't one ready yet"""
    try:
        deployments = load_v2_deployments_json(service=service, soa_dir=soa_dir)
        return deployments.get_deployment_version_for_deploy_group(
            deploy_group=deploy_group
        )
    except NoDeploymentsAvailable:
        return None



import json
import logging
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import TypedDict
from typing import Union

import service_configuration_lib
from kubernetes.client import ApiClient

from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfigDict
from paasta_tools.kubernetes_tools import sanitised_cr_name
from paasta_tools.long_running_service_tools import load_service_namespace_config
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_git_sha_from_dockerurl
from paasta_tools.utils import load_service_instance_config
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import load_v2_deployments_json


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


KUBERNETES_NAMESPACE = "paasta-vitessclusters"


# Global variables
TOPO_IMPLEMENTATION = "zk2"
TOPO_GLOBAL_ROOT = "/vitess-paasta/global"
WEB_PORT = "15000"
GRPC_PORT = "15999"


# Environment variables
VTCTLD_EXTRA_ENV = {
    "WEB_PORT": WEB_PORT,
    "GRPC_PORT": GRPC_PORT,
}


# Extra Flags
VTADMIN_EXTRA_FLAGS = {"grpc-allow-reflection": "true"}

VTCTLD_EXTRA_FLAGS = {
    "disable_active_reparents": "true",
    "security_policy": "read-only",
}


class KVEnvVar(TypedDict, total=False):
    name: str
    value: str


class KVEnvVarValueFrom(TypedDict, total=False):
    name: str
    valueFrom: Dict[str, Any]


class RequestsDict(TypedDict, total=False):
    cpu: str
    memory: str
    disk: str


class ResourceConfigDict(TypedDict, total=False):
    replicas: int
    requests: Dict[str, RequestsDict]
    limits: Dict[str, RequestsDict]


class GatewayConfigDict(TypedDict, total=False):
    replicas: int


class CellConfigDict(TypedDict, total=False):
    name: str
    gateway: GatewayConfigDict


class VitessDashboardConfigDict(TypedDict, total=False):
    cells: List[str]
    affinity: Dict[str, Any]
    extraEnv: List[Union[KVEnvVar, KVEnvVarValueFrom]]
    extraFlags: Dict[str, str]
    extraLabels: Dict[str, str]
    replicas: int
    resources: Dict[str, Any]
    annotations: Mapping[str, Any]


class VtAdminConfigDict(TypedDict, total=False):
    cells: List[str]
    apiAddresses: List[str]
    affinity: Dict[str, Any]
    extraEnv: List[Union[KVEnvVar, KVEnvVarValueFrom]]
    extraFlags: Dict[str, str]
    extraLabels: Dict[str, str]
    replicas: int
    readOnly: bool
    apiResources: Dict[str, Any]
    webResources: Dict[str, Any]
    annotations: Mapping[str, Any]


def get_formatted_environment_variables(
    env_vars: Dict[str, Any]
) -> List[Union[KVEnvVar, KVEnvVarValueFrom]]:
    """
    Helper function to take in key value pairs of environment variables and return a list of dicts
    """
    updated_environment_variables: List[Union[KVEnvVar, KVEnvVarValueFrom]] = []
    for env_key, env_value in env_vars.items():
        if isinstance(env_value, str):
            updated_environment_variables.append(
                KVEnvVar(name=env_key, value=env_value)
            )
        elif isinstance(env_value, dict):
            updated_environment_variables.append(
                KVEnvVarValueFrom(name=env_key, valueFrom=env_value)
            )
        else:
            log.error(f"Invalid environment variable {env_key}={env_value}")
    return updated_environment_variables


def get_cell_config(
    cell: str,
) -> CellConfigDict:
    """
    get vtgate config
    """
    config = CellConfigDict(
        name=cell,
        gateway=GatewayConfigDict(
            replicas=0,
        ),
    )
    return config


def get_vitess_dashboard_config(
    cells: List[str],
    zk_address: str,
    vtctld_resources: ResourceConfigDict,
    env: List[Union[KVEnvVar, KVEnvVarValueFrom]],
    labels: Dict[str, str],
    node_affinity: dict,
    annotations: Mapping[str, Any],
) -> VitessDashboardConfigDict:
    """
    get vtctld config
    """
    replicas = vtctld_resources.get("replicas")
    requests = vtctld_resources.get(
        "requests", RequestsDict(cpu="100m", memory="256Mi")
    )
    environment_overrides: Dict[str, Any] = {
        "TOPOLOGY_FLAGS": f"--topo_implementation {TOPO_IMPLEMENTATION} --topo_global_server_address {zk_address} --topo_global_root {TOPO_GLOBAL_ROOT}",
    }
    environment_overrides.update(VTCTLD_EXTRA_ENV)
    updated_vtctld_extra_env = (
        get_formatted_environment_variables(environment_overrides) + env
    )

    config = VitessDashboardConfigDict(
        cells=cells,
        affinity={"nodeAffinity": node_affinity},
        extraEnv=updated_vtctld_extra_env,
        extraFlags=VTCTLD_EXTRA_FLAGS,
        extraLabels=labels,
        replicas=replicas,
        resources={
            "requests": requests,
            "limits": requests,
        },
        annotations=annotations,
    )
    return config


def get_vt_admin_config(
    cells: List[str],
    vtadmin_resources: ResourceConfigDict,
    env: List[Union[KVEnvVar, KVEnvVarValueFrom]],
    labels: Dict[str, str],
    node_affinity: dict,
    annotations: Mapping[str, Any],
) -> VtAdminConfigDict:
    """
    get vtadmin config
    """
    replicas = vtadmin_resources.get("replicas")
    requests = vtadmin_resources.get(
        "requests", RequestsDict(cpu="100m", memory="256Mi")
    )
    config = VtAdminConfigDict(
        cells=cells,
        apiAddresses=["http://localhost:15000"],
        affinity={"nodeAffinity": node_affinity},
        extraLabels=labels,
        extraFlags=VTADMIN_EXTRA_FLAGS,
        extraEnv=env,
        replicas=replicas,
        readOnly=False,
        apiResources={
            "requests": requests,
            "limits": requests,
        },
        webResources={
            "requests": requests,
            "limits": requests,
        },
        annotations=annotations,
    )
    return config


class VitessDeploymentConfigDict(KubernetesDeploymentConfigDict, total=False):
    images: Dict[str, str]
    zk_address: str


class VitessClusterConfigDict(VitessDeploymentConfigDict, total=False):
    cells: List[CellConfigDict]
    vitessDashboard: VitessDashboardConfigDict
    vtadmin: VtAdminConfigDict
    updateStrategy: Dict[str, str]
    globalLockserver: Dict[str, Dict[str, str]]


class VitessClusterInstanceConfigDict(KubernetesDeploymentConfigDict, total=False):
    cells: List[str]
    zk_address: str
    vtctld_resources: ResourceConfigDict
    vtgate_resources: ResourceConfigDict
    vttablet_resources: ResourceConfigDict
    vtadmin_resources: ResourceConfigDict
    images: Dict[str, str]


class VitessDeploymentConfig(KubernetesDeploymentConfig):
    def get_namespace(self) -> str:
        return KUBERNETES_NAMESPACE

    def get_env_variables(self) -> List[Union[KVEnvVar, KVEnvVarValueFrom]]:
        # get all K8s container env vars and format their keys to camel case

        # Workaround from https://github.com/kubernetes-client/python/issues/390
        api_client = ApiClient()
        env = [
            api_client.sanitize_for_serialization(env)
            for env in self.get_container_env()
        ]
        return env

    def get_labels(self) -> Dict[str, str]:
        # get default labels from parent class to adhere to paasta contract
        docker_url = self.get_docker_url(
            system_paasta_config=load_system_paasta_config()
        )
        git_sha = get_git_sha_from_dockerurl(docker_url)
        labels = self.get_kubernetes_metadata(git_sha=git_sha).labels
        if "yelp.com/owner" in labels.keys():
            labels["yelp.com/owner"] = "dre_mysql"
        return labels

    def get_annotations(self) -> Mapping[str, Any]:
        # get required annotations to be added to the formatted resource before creating or updating custom resource
        service_namespace_config = load_service_namespace_config(
            service=self.service, namespace=self.get_nerve_namespace()
        )
        system_paasta_config = load_system_paasta_config()
        has_routable_ip = self.has_routable_ip(
            service_namespace_config, system_paasta_config
        )
        annotations: Mapping[str, Any] = {
            "smartstack_registrations": json.dumps(self.get_registrations()),
            "paasta.yelp.com/routable_ip": has_routable_ip,
        }

        return annotations

    def get_vitess_node_affinity(self) -> dict:
        # Workaround from https://github.com/kubernetes-client/python/issues/390
        api_client = ApiClient()
        node_affinity = api_client.sanitize_for_serialization(self.get_node_affinity())
        return node_affinity

    def get_region(self) -> str:
        superregion = self.get_cluster()
        superregion_to_region_map = (
            load_system_paasta_config().get_superregion_to_region_mapping()
        )
        region = None
        for superregion_prefix in superregion_to_region_map:
            if superregion.startswith(superregion_prefix):
                region = superregion.replace(
                    superregion_prefix, superregion_to_region_map[superregion_prefix]
                )
        if region is None:
            log.error(
                f"Region not found for superregion {superregion}. Check superregion_to_region_mapping in system paasta config"
            )
            # Exiting early here since region is needed to fetch secrets from vault
            sys.exit(1)
        return region

    def get_update_strategy(self) -> Dict[str, str]:
        return {"type": "Immediate"}


class VitessClusterConfig(VitessDeploymentConfig):
    config_dict: VitessClusterInstanceConfigDict

    config_filename_prefix = "vitesscluster"

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: VitessClusterConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,  # superregion
            instance=instance,  # host-1
            service=service,  # vitess
            soa_dir=soa_dir,
            config_dict=config_dict,
            branch_dict=branch_dict,
        )

    def get_images(self) -> Dict[str, str]:
        vitess_images = self.config_dict.get(
            "images", load_system_paasta_config().get_vitess_images()
        )
        return {
            "vtctld": vitess_images["vtctld_image"],
            "vtadmin": vitess_images["vtadmin_image"],
            "vtgate": vitess_images["vtgate_image"],
            "vttablet": vitess_images["vttablet_image"],
        }

    def get_global_lock_server(self) -> Dict[str, Dict[str, str]]:
        zk_address = self.config_dict.get("zk_address")
        return {
            "external": {
                "implementation": TOPO_IMPLEMENTATION,
                "address": zk_address,
                "rootPath": TOPO_GLOBAL_ROOT,
            }
        }

    def get_vitess_dashboard(self) -> VitessDashboardConfigDict:
        cells = self.config_dict.get("cells")
        zk_address = self.config_dict.get("zk_address")
        vtctld_resources = self.config_dict.get("vtctld_resources")

        formatted_env = self.get_env_variables()
        labels = self.get_labels()
        node_affinity = self.get_vitess_node_affinity()
        annotations = self.get_annotations()

        return get_vitess_dashboard_config(
            cells,
            zk_address,
            vtctld_resources,
            formatted_env,
            labels,
            node_affinity,
            annotations,
        )

    def get_vtadmin(self) -> VtAdminConfigDict:
        cells = self.config_dict.get("cells")
        vtadmin_resources = self.config_dict.get("vtadmin_resources")

        formatted_env = self.get_env_variables()
        labels = self.get_labels()
        node_affinity = self.get_vitess_node_affinity()
        annotations = self.get_annotations()

        return get_vt_admin_config(
            cells, vtadmin_resources, formatted_env, labels, node_affinity, annotations
        )

    def get_cells(self) -> List[CellConfigDict]:
        cells = self.config_dict.get("cells")
        return [get_cell_config(cell) for cell in cells]

    def get_vitess_config(self) -> VitessClusterConfigDict:
        vitess_config = VitessClusterConfigDict(
            namespace=self.get_namespace(),
            images=self.get_images(),
            globalLockserver=self.get_global_lock_server(),
            cells=self.get_cells(),
            vitessDashboard=self.get_vitess_dashboard(),
            vtadmin=self.get_vtadmin(),
            updateStrategy=self.get_update_strategy(),
        )
        return vitess_config

    def validate(
        self,
        params: List[str] = [
            "cpus",
            "security",
            "dependencies_reference",
            "deploy_group",
        ],
    ) -> List[str]:
        # Use InstanceConfig to validate shared config keys like cpus and mem
        # TODO: add mem back to this list once we fix PAASTA-15582 and
        # move to using the same units as flink/marathon etc.
        error_msgs = super().validate(params=params)

        if error_msgs:
            name = self.get_instance()
            return [f"{name}: {msg}" for msg in error_msgs]
        else:
            return []


def load_vitess_cluster_instance_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessClusterConfig:
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    instance_config = load_service_instance_config(
        service, instance, "vitesscluster", cluster, soa_dir=soa_dir
    )
    general_config = deep_merge_dictionaries(
        overrides=instance_config, defaults=general_config
    )

    branch_dict: Optional[BranchDictV2] = None
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = VitessClusterConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    vitess_cluster_config = VitessClusterConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )

    return vitess_cluster_config


def load_vitess_cluster_instance_configs(
    service: str,
    instance: str,
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessClusterConfigDict:
    vitess_cluster_instance_configs = load_vitess_cluster_instance_config(
        service, instance, cluster, soa_dir=soa_dir
    ).get_vitess_config()
    return vitess_cluster_instance_configs


# TODO: read this from CRD in service configs
def cr_id(service: str, instance: str) -> Mapping[str, str]:
    return dict(
        group="planetscale.com",
        version="v2",
        namespace=KUBERNETES_NAMESPACE,
        plural="vitessclusters",
        name=sanitised_cr_name(service, instance),
    )



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import TypedDict
from typing import Union

import service_configuration_lib

from paasta_tools.kubernetes_tools import KubernetesDeploymentConfigDict
from paasta_tools.kubernetes_tools import limit_size_with_hash
from paasta_tools.kubernetes_tools import sanitised_cr_name
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_service_instance_config
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.vitesscluster_tools import get_formatted_environment_variables
from paasta_tools.vitesscluster_tools import KVEnvVar
from paasta_tools.vitesscluster_tools import KVEnvVarValueFrom
from paasta_tools.vitesscluster_tools import RequestsDict
from paasta_tools.vitesscluster_tools import ResourceConfigDict
from paasta_tools.vitesscluster_tools import VitessDeploymentConfig
from paasta_tools.vitesscluster_tools import VitessDeploymentConfigDict


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


KUBERNETES_NAMESPACE = "paasta-vitessclusters"


# Global variables
TOPO_IMPLEMENTATION = "zk2"
TOPO_GLOBAL_ROOT = "/vitess-paasta/global"
WEB_PORT = "15000"
GRPC_PORT = "15999"


# Environment variables
VTTABLET_EXTRA_ENV = {
    "WEB_PORT": WEB_PORT,
    "GRPC_PORT": GRPC_PORT,
    "SHARD": "0",
    "EXTERNAL_DB": "1",
    "ROLE": "replica",
    "VAULT_ROLEID": {
        "secretKeyRef": {
            "name": "paasta-vitessclusters-secret-vitess-k8s-vault-vttablet-approle-roleid",
            "key": "vault-vttablet-approle-roleid",
        }
    },
    "VAULT_SECRETID": {
        "secretKeyRef": {
            "name": "paasta-vitessclusters-secret-vitess-k8s-vault-vttablet-approle-secretid",
            "key": "vault-vttablet-approle-secretid",
        }
    },
}

# Extra Flags
VTTABLET_EXTRA_FLAGS = {
    "log_err_stacks": "true",
    "grpc_max_message_size": "134217728",
    "init_tablet_type": "replica",
    "queryserver-config-schema-reload-time": "1800",
    "dba_pool_size": "4",
    "vreplication_heartbeat_update_interval": "60",
    "vreplication_tablet_type": "REPLICA",
    "keep_logs": "72h",
    "enable-lag-throttler": "true",
    "throttle_check_as_check_self": "true",
    "db_charset": "utf8mb4",
    "disable_active_reparents": "true",
}


class VtTabletDict(TypedDict, total=False):
    extraFlags: Dict[str, str]
    resources: Dict[str, Any]


class TabletPoolDict(TypedDict, total=False):
    cell: str
    name: str
    type: str
    affinity: Dict[str, Any]
    extraLabels: Dict[str, str]
    extraEnv: List[Union[KVEnvVar, KVEnvVarValueFrom]]
    extraVolumeMounts: List[Dict[str, Any]]
    extraVolumes: List[Dict[str, Any]]
    replicas: int
    vttablet: VtTabletDict
    externalDatastore: Dict[str, Any]
    dataVolumeClaimTemplate: Dict[str, Any]
    annotations: Mapping[str, Any]


class ShardTemplateDict(TypedDict, total=False):
    databaseInitScriptSecret: Dict[str, str]
    tabletPools: List[TabletPoolDict]


class PartitioningValueDict(TypedDict, total=False):
    parts: int
    shardTemplate: ShardTemplateDict


class KeyspaceConfigDict(TypedDict, total=False):
    durabilityPolicy: str
    turndownPolicy: str
    partitionings: List[Dict[str, PartitioningValueDict]]
    name: str


def get_tablet_pool_config(
    cell: str,
    db_name: str,
    keyspace: str,
    host: str,
    zk_address: str,
    throttle_query_table: str,
    throttle_metrics_threshold: str,
    tablet_type: str,
    region: str,
    vttablet_resources: ResourceConfigDict,
    env: List[Union[KVEnvVar, KVEnvVarValueFrom]],
    labels: Dict[str, str],
    node_affinity: dict,
    annotations: Mapping[str, Any],
) -> TabletPoolDict:
    """
    get vttablet config
    """
    vttablet_extra_flags = VTTABLET_EXTRA_FLAGS.copy()
    flag_overrides = {
        "throttle_metrics_query": f"select max_replication_delay from max_mysql_replication_delay.{throttle_query_table};",
        "throttle_metrics_threshold": throttle_metrics_threshold,
        "enforce-tableacl-config": "true",
        "table-acl-config": f"/nail/srv/configs/vitess_keyspace_acls/acls_for_{db_name}.json",
        "table-acl-config-reload-interval": "60s",
        "queryserver-config-strict-table-acl": "true",
        "db-credentials-server": "vault",
        "db-credentials-vault-addr": f"https://vault-dre.{region}.yelpcorp.com:8200",
        "db-credentials-vault-path": "secrets/vitess/vt-tablet/vttablet_credentials.json",
        "db-credentials-vault-tls-ca": f"/etc/vault/all_cas/acm-privateca-{region}.crt",
        "db-credentials-vault-ttl": "60s",
    }
    vttablet_extra_flags.update(flag_overrides)

    environment_overrides: Dict[str, Any] = {
        "VAULT_ADDR": f"https://vault-dre.{region}.yelpcorp.com:8200",
        "VAULT_CACERT": f"/etc/vault/all_cas/acm-privateca-{region}.crt",
        "TOPOLOGY_FLAGS": f"--topo_implementation {TOPO_IMPLEMENTATION} --topo_global_server_address ${zk_address} --topo_global_root {TOPO_GLOBAL_ROOT}",
        "CELL_TOPOLOGY_SERVERS": zk_address,
        "DB": db_name,
        "KEYSPACE": keyspace,
    }
    environment_overrides.update(VTTABLET_EXTRA_ENV)
    updated_vttablet_extra_env = (
        get_formatted_environment_variables(environment_overrides) + env
    )

    # Add extra pod label to filter
    tablet_type_label = limit_size_with_hash(name=f"{db_name}_{tablet_type}", limit=63)
    labels.update({"tablet_type": tablet_type_label})

    try:
        type = load_system_paasta_config().get_vitess_tablet_pool_type_mapping()[
            tablet_type
        ]
    except KeyError:
        log.error(
            f"Tablet type {tablet_type} not found in system paasta config vitess_tablet_pool_type_mapping"
        )
        type = "externalmaster"

    replicas = vttablet_resources.get("replicas")
    requests = vttablet_resources.get(
        "requests", RequestsDict(cpu="100m", memory="256Mi")
    )

    config = TabletPoolDict(
        cell=cell,
        name=f"{db_name}_{tablet_type}",
        type=type,
        affinity={"nodeAffinity": node_affinity},
        extraLabels=labels,
        extraEnv=updated_vttablet_extra_env,
        extraVolumeMounts=[
            {
                "mountPath": "/etc/vault/all_cas",
                "name": "vault-secrets",
                "readOnly": True,
            },
            {
                "mountPath": "/nail/srv",
                "name": "srv-configs",
                "readOnly": True,
            },
            {
                "mountPath": "/nail/etc/srv-configs",
                "name": "etc-srv-configs",
                "readOnly": True,
            },
            {
                "mountPath": "etc/credentials.yaml",
                "name": "vttablet-fake-credentials",
                "readOnly": True,
            },
            {
                "mountPath": "/etc/init_db.sql",
                "name": "keyspace-fake-init-script",
                "readOnly": True,
            },
        ],
        extraVolumes=[
            {"name": "vault-secrets", "hostPath": {"path": "/nail/etc/vault/all_cas"}},
            {
                "name": "srv-configs",
                "hostPath": {"path": "/nail/srv"},
            },
            {
                "name": "etc-srv-configs",
                "hostPath": {"path": "/nail/etc/srv-configs"},
            },
            {"name": "vttablet-fake-credentials", "hostPath": {"path": "/dev/null"}},
            {"name": "keyspace-fake-init-script", "hostPath": {"path": "/dev/null"}},
        ],
        replicas=replicas,
        vttablet={
            "extraFlags": vttablet_extra_flags,
            "resources": {
                "requests": requests,
                "limits": requests,
            },
        },
        externalDatastore={
            "database": db_name,
            "host": host,
            "port": 3306,
            "user": "vt_app",
            "credentialsSecret": {
                "key": "/etc/credentials.yaml",
                "volumeName": "vttablet-fake-credentials",
            },
        },
        dataVolumeClaimTemplate={
            "accessModes": ["ReadWriteOnce"],
            "resources": {"requests": {"storage": "10Gi"}},
            "storageClassName": "ebs-csi-gp3",
        },
        annotations=annotations,
    )
    return config


class VitessKeyspaceConfigDict(VitessDeploymentConfigDict, total=False):
    name: str
    databaseName: str
    durabilityPolicy: str
    turndownPolicy: str
    partitionings: List[Dict[str, PartitioningValueDict]]
    updateStrategy: Dict[str, str]
    globalLockserver: Dict[str, str]
    zoneMap: Dict[str, Any]


def get_keyspace_config(
    cells: List[str],
    keyspace: str,
    cluster: str,
    vttablet_resources: ResourceConfigDict,
    images: Dict[str, str],
    update_strategy: Dict[str, str],
    global_lockserver: Dict[str, str],
    zk_address: str,
    region: str,
    env: List[Union[KVEnvVar, KVEnvVarValueFrom]],
    labels: Dict[str, str],
    node_affinity: dict,
    annotations: Mapping[str, Any],
) -> VitessKeyspaceConfigDict:
    """
    get vitess keyspace config
    """
    db_name = keyspace

    tablet_pools = []

    # get vttablets
    tablet_types = load_system_paasta_config().get_vitess_tablet_types()
    for tablet_type in tablet_types:
        ecosystem = region.split("-")[-1]
        host = f"mysql-{cluster}-{tablet_type}.dre-{ecosystem}"

        # We use migration_replication delay for migration tablets and read_replication_delay for everything else
        # Also throttling threshold for refresh and sanitized primaries is set at 30 seconds and everything else at 3 seconds
        try:
            throttling_configs = (
                load_system_paasta_config().get_vitess_throttling_config()
            )
            throttle_query_table = throttling_configs[tablet_type][
                "throttle_query_table"
            ]
            throttle_metrics_threshold = throttling_configs[tablet_type][
                "throttle_metrics_threshold"
            ]
        except KeyError:
            log.error(
                f"Throttling configs for tablet type {tablet_type} not found in system paasta config vitess_throttling_configs"
            )

        if cluster.startswith("refresh") or cluster.startswith("sanitized"):
            throttle_metrics_threshold = "30"
        else:
            throttle_metrics_threshold = "3"

        tablet_pools.extend(
            [
                get_tablet_pool_config(
                    cell,
                    db_name,
                    keyspace,
                    host,
                    zk_address,
                    throttle_query_table,
                    throttle_metrics_threshold,
                    tablet_type,
                    region,
                    vttablet_resources,
                    env,
                    labels,
                    node_affinity,
                    annotations,
                )
                for cell in cells
            ]
        )
    vitess_keyspace_config = VitessKeyspaceConfigDict(
        name=keyspace,
        durabilityPolicy="none",
        turndownPolicy="Immediate",
        images=images,
        databaseName=db_name,
        updateStrategy=update_strategy,
        globalLockserver=global_lockserver,
        partitionings=[
            {
                "equal": PartitioningValueDict(
                    parts=1,
                    shardTemplate=ShardTemplateDict(
                        databaseInitScriptSecret={
                            "volumeName": "keyspace-fake-init-script",
                            "key": "/etc/init_db.sql",
                        },
                        tabletPools=tablet_pools,
                    ),
                )
            }
        ],
        zoneMap={},
    )
    return vitess_keyspace_config


class VitessKeyspaceInstanceConfigDict(KubernetesDeploymentConfigDict, total=False):
    cells: List[str]
    zk_address: str
    images: Dict[str, str]
    keyspace: str
    cluster: str
    vttablet_resources: ResourceConfigDict


class VitessKeyspaceConfig(VitessDeploymentConfig):
    config_dict: VitessKeyspaceInstanceConfigDict

    config_filename_prefix = "vitesskeyspace"

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: VitessKeyspaceConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,  # superregion
            instance=instance,  # host-1
            service=service,  # vitess
            soa_dir=soa_dir,
            config_dict=config_dict,
            branch_dict=branch_dict,
        )

    def get_images(self) -> Dict[str, str]:
        vitess_images = self.config_dict.get(
            "images", load_system_paasta_config().get_vitess_images()
        )
        return {
            "vtctld": vitess_images["vtctld_image"],
            "vtadmin": vitess_images["vtadmin_image"],
            "vtgate": vitess_images["vtgate_image"],
            "vttablet": vitess_images["vttablet_image"],
        }

    def get_global_lock_server(self) -> Dict[str, str]:
        zk_address = self.config_dict.get("zk_address")
        return {
            "implementation": TOPO_IMPLEMENTATION,
            "address": zk_address,
            "rootPath": TOPO_GLOBAL_ROOT,
        }

    def get_vitess_config(self) -> VitessKeyspaceConfigDict:
        cells = self.config_dict.get("cells")
        zk_address = self.config_dict.get("zk_address")
        region = self.get_region()
        keyspace = self.config_dict.get("keyspace")
        cluster = self.config_dict.get("cluster")
        vttablet_resources = self.config_dict.get("vttablet_resources")
        global_lockserver = self.get_global_lock_server()
        images = self.get_images()
        update_strategy = self.get_update_strategy()

        formatted_env = self.get_env_variables()
        labels = self.get_labels()
        node_affinity = self.get_vitess_node_affinity()
        annotations = self.get_annotations()

        return get_keyspace_config(
            cells,
            keyspace,
            cluster,
            vttablet_resources,
            images,
            update_strategy,
            global_lockserver,
            zk_address,
            region,
            formatted_env,
            labels,
            node_affinity,
            annotations,
        )

    def validate(
        self,
        params: List[str] = [
            "cpus",
            "security",
            "dependencies_reference",
            "deploy_group",
        ],
    ) -> List[str]:
        # Use InstanceConfig to validate shared config keys like cpus and mem
        # TODO: add mem back to this list once we fix PAASTA-15582 and
        # move to using the same units as flink/marathon etc.
        error_msgs = super().validate(params=params)

        if error_msgs:
            name = self.get_instance()
            return [f"{name}: {msg}" for msg in error_msgs]
        else:
            return []


def load_vitess_keyspace_instance_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessKeyspaceConfig:
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    instance_config = load_service_instance_config(
        service, instance, "vitesskeyspace", cluster, soa_dir=soa_dir
    )
    general_config = deep_merge_dictionaries(
        overrides=instance_config, defaults=general_config
    )

    branch_dict: Optional[BranchDictV2] = None
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = VitessKeyspaceConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    vitess_keyspace_config = VitessKeyspaceConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )

    return vitess_keyspace_config


def load_vitess_keyspace_instance_configs(
    service: str,
    instance: str,
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessKeyspaceConfigDict:
    vitess_keyspace_instance_configs = load_vitess_keyspace_instance_config(
        service, instance, cluster, soa_dir=soa_dir
    ).get_vitess_config()
    return vitess_keyspace_instance_configs


# TODO: read this from CRD in service configs
def cr_id(service: str, instance: str) -> Mapping[str, str]:
    return dict(
        group="planetscale.com",
        version="v2",
        namespace=KUBERNETES_NAMESPACE,
        plural="vitesskeyspaces",
        name=sanitised_cr_name(service, instance),
    )
